#!/usr/bin/perl 
# P855/860 disk premark utility version 0.5 -- date July 1 2008
# Creates/Formats an empty disk image for the P855/P860 emulator.
#
# The P855 and P860 supported various fixed head disks and a moving head disk, 
# the X1210. This utility is primarily meant for creating X1210 diskpack images 
# (and for X1215 diskpack images supported later by the P860 and other P800 
# models, and which is compatible with the X1210 diskpack).
#
# The X1210 was a moving head disk, with a removable cardridge with a single
# platter, supporting 2 moving heads, one for side 0 and one for side 1.
# Each side supported 203 tracks, with 16 sectors of 200 user data words per track.
# The disk unit was supported via the multiplex channel, where a disk control 
# unit was able to support 2 disk units
#
# Usage: premark.pl <disk image filename>.img
#
# premark uses ipl2disk.pl to load the disk ipl progarm in physical sector 3
# of the premarked disk.
#
# REMARKS / TODO
# - IPL in sector 1 has to be added (solved in version 0.4)
# - the IPL program as used by the original P800 system is unknown; However the
#   bootstrap to load and start the IPL program is known.
# - the format of the volume label data, the offsets of the volume label and 
#   the granule table as used by the original P800 software are unknown; values 
#   are taken, which might not be compatible with the origional P800 system.
#   >>>>>> Solved in version 0.5
# - not known is whether allocated granules in the granule table are indicated
#   with a 1 or a 0: we assume in this program that a 1 indicates an 
#   allocated granule. This assumption is wrong: 1 indicates a free granule
#   >>>>>> Adapted in version 0.5
# - the sector identifier (the 1st word with the 205 word sector) is said
#   to contain the cylinder number in which the sector is located. Unknown is
#   where the cylinder number is located in the 16 bit word. In this program
#   it is assumed the the cylinder number is located in the least significant 
#   bits if the 16 bit word, which is ok.
#   For disk integrity tests/debugging this program also adds the physical
#   sector number to the sector identifier (the 5 most significant bits in the
#   16 bit word). Must perhaps be removed in future!
#   In version 0.5 the physical sector number is removed from the sector identifier.
# - also, not compatible with the original p800 disk premark is that this one
#   reserves an area on disk to store the monitor load module (utility mon2disk).
# . The monitor must be stored in consecutive granules, starting with granule 2. 
#   8 granules are reserved, allowing for a monitor with a length of 11.6 kw maximum.

# mofifications:
# version 0.5
# - BITTAB changed: free granules are indicated with a 1 and allocated granules
#   with a zero
# - sector ident changed conform the original diskpacks: only cylinder number,
#   right justified in word 0 of the sector.
# - Volume label corrected conform original disk image
# - Catalogue and Directory corrected

use strict;
use bytes;
my $DEBUG = 1;
use constant P800 => 1;
use constant TEST => 2;

my $version = 0.5;
###############################################################
# disk pack parameters
my $type  = 'XM'; # disk unit type
my $cyls  = 203;  # number of tracks per side = number of cylinders
my $heads = 2;    # number of heads
my $spt   = 16;   # number of sectors per track
my $lsec  = 205;  # sector length is 205 16 bit words
#
my $cyl;          # working variable
my $head;         # working variable
my $sector;       # working variable
my $iplfile =     "IPL.abs";
# select 1 of 2:
#my $SEC_IDENT=TEST;  # generates a test sector identifier which includes the physical sector number
my $SEC_IDENT=P800; # generates the original sector identifier
###############################################################
# Each sector consists of 205 16 bit data words, 5 overhead and 200 for user data.
# Sector format:
use constant IDENT => 0x0; # word 1 is the sector identifier
use constant SEQM  => 0x2; # word 2 is for sequential file admin
use constant REC   => 0x4; # word 3 is start of the data part
#   xxxxxxxx 200 data words
#   xxxxxxxx 3   reserved words
#
# Word-1, the sector identifier contained the cylinder number to check the seek 
# operation.
# Of all known disk units supported by the P800 family of minicomputers, the
# maximum number of cylinders was 512, so this identifier needs 9 bits.
# The other 7 bits in this simulator environment are used for debugging
# to contain head and logical sector number:
#
# Physical sectors in a track are consecutive, logical sectors not.
# The system software uses always logical sectors; mapping in a track is
# as follows (for the X1210/X1215):
#
# physical sector  logical sector
#    sector 0   <->   sector 0
#    sector 1   <->   sector 11
#    sector 2   <->   sector 6
#    sector 3   <->   sector 1
#    sector 4   <->   sector 12
#    sector 5   <->   sector 7
#    sector 6   <->   sector 2
#    sector 7   <->   sector 13
#    sector 8   <->   sector 8
#    sector 9   <->   sector 3
#    sector 10  <->   sector 14
#    sector 11  <->   sector 9
#    sector 12  <->   sector 4
#    sector 13  <->   sector 15
#    sector 14  <->   sector 10
#    sector 15  <->   sector 5
#
# physical-sector-in-track = remainder( modulo-16 (logical-sector-in-track * 3))
#
# for the X1210/X1215 the format of the sector identifier is chosen for
# testing as follows (selected with SEC_IDENT=TEST):
# x  xxxx  00 xxxxxxxxx
# ^     ^             ^
# head  physsector    cylinder
# 0/1   0..15         0..202
#
# The original sector identifier is formatted as follows:
# 0000 000x xxxx xxxx
#                   ^
#            cylinder
#              0..202
# This format is selected with: SEC_IDENT=P800
#
#
#
# Word-2, used for sequential files, is formatted as follows:
# - status: bits 0,1,2
# - length: bits 7..15
# status bit 0 = 1 means: sector has been deleted
# status bit 1 = 1 means: an EOS(End Of Segment) has been written in this sector.
#                         (The EOS record is the last record of the file written
#                          in this sector. The 1st record following the EOS  
#                          always starts in a new sector)
# status bit 2 = 1 means: an EOF(End Of File) record has been written in this
#                         sector. The EOF record requires a full sector without
#                         any other records written in it.
# length (bits 7..15) means: length of the data part written in this sector
#                         in characters (up to 400)
# The data area in the sector contains the data written either in sequential
# mode or in random access mode. When data is written in random access mode,
# then word-2 remains unused.
#
# Disk space allocation
# Space on disk is allocated dynamically. Space on disk is organized therefore
# in socalled granules, where each granule is 8 consecutive (logical) sectors. 
# The granule is the unit of space allocation to a file. Space allocation to a 
# file is maintained in an allocation table of 200 entries in the file, one 
# entry for each allocated granule. This limits the maximum file length to 
# 200*8 sectors of 400 bytes = 640 kbytes. The allocation table takes 
# sector 1 of the file (sector 0 is unused by the system)
#
my $n = $cyls*$heads*$spt;                # number of sectors on disk
my $g = int($n/8);                        # number of granules

# An X1210 disk contains 203 * 2 * 16 = 6496 sectors
# An X1210 disk contains 6496/8       = 812  granules
#
# In the global disk space allocation table, each granule is represented by a
# bit. For the X1210, 102 bytes (= 816 bits) are required to administrate
# the 812 granules (102 bytes = 66 hex bytes).
# The disk allocation table is located in sector 0 of the disk, together
# with the volume administration data of the volume: the "volume label"
#
# Volume label
# ============
# The volume label starts directly after the sector header so at offset 4
use constant VOLLAB  => 4;                # volume label start
my $vls1= "    LABEL = ";                 # l=12 bytes
use constant LABEL   => 16;               # offset to the label field: 16 bytes long
use constant VOLDAT  => 32;               # date string offset (date=premark date)
my $vls2= "  DATE =  ";                   # l=10 bytes
use constant DAY     => 42;               # offset to day field: 2 bytes long
my $vls3= "  ";                           # 2 bytes (spaces)
use constant MONTH   => 46;               # offset to monthfield: 2 bytes long
my $vls4= "  ";                           # 2 bytes (spaces)
use constant YEAR    => 50;               # offset to yearfield: 2 bytes long
my $vls5= "  1215PACK NBR =  ";           # l=18 bytes
use constant PACKNBR => 70;               # pack number field: 4 bytes (pack numer in ascii)
use constant BADTRKL => 74;               # offset to badtracklist (2bytes: pointer to physical sector)
use constant SECPTRK => 76;               # offset to sectors/track (2 bytes long)
use constant INTERLC => 78;               # offset to number of interlaces (2 bytes long)
use constant SECSIZE => 80;               # sector size in chars (incl header/trailer words) (2 bytes long)
use constant VTOC    => 82;               # offset to physical sector number of VTOC (2 bytes long)
# BITTAB
# After the volume label comes the BITTAB: the granule allocation table.
use constant BITLEN  => 84;               # offset to the length of BITTAB (length in bytes)
use constant BITTAB  => 86;               # offset granule allocation table in sector 0
use constant BITTABL => 102;              # length of allocation table (102 bytes)
my $label   = "";                         # disk pack label (16 chars)
my $date    = "";                         # format date (8 chars)
my $day     = "";                         # 2 chars
my $month   = "";                         # 2 chars
my $year    = "";                         # 2 chars
my $packnbr = "";                         # pack number (4 digits)
my $sysid   = "";                         # 8 chars

# Disk layout
# ===========
# granule 0
#   sector0 = volume label/granule allocation table
#   sector1 = IPL
#   sector2..5 = catalogue (6=badtracklist, 7=VTOC not initiated at thsi moment)
#      each catalogue entry takes 8 16-bit words
#      maximum = 4 * 25 = 100 catelogue entries (each entry pointing to a directory)
#      Entry 0 is just CATALOGUE
#      The 1st catalogue entry is for the system and it points to the system 
#      directory with system software.
#      The catalogue ends with an all one 16 bit word (0xFFFF)
# granule 1 = system directory
#      each directory entry takes 8 16-bit words
#      maximum = 8 * 25 = 200 directory entries (each entry pointing to a file)
#      The system directory starts at (logical) sector 8
#      The 1st directory entry is the monitor (sector 16)
#      The directory ends with an all one 16 bit word (0xFFFF)
#      User directories use the same format
# granule 2..811 = monitor, user directories and files
#      the monitor file takes consecutive granules 2..n (on a system disk)
#      the monitor load module starts at sector 16
# NOTE:on the P800 system, the monitor was stored on the diskpack during
#      system generation, which also allocated/reserved space on the disk
#      pack for that monitor. In this PREMARK we use a different approach:
#      just 8 granules (granule 2..9, ~ 11.6kwords) are reseved for storing a 
#      monitor load module file, starting at granule 2 (logical sector 16).
#      This monitor is meant to be loaded and started by the IPL in sector 1.
#
# Initial Granule Table (102 bytes)
# a 1 bit indicates an allocated granule, 0 a free one.
#
#   xxxx xxxx < 1st byte of disk granule space allocation table
#   ^       ^
#   ^       granule 7
#   ^
#   granule 0
#
my $gr0 = 0x00;   # w1      10 granules reserved; see above
my $gr1 = 0x3F;   # w2      label.., system dir, 8 for a monitor
my $gr2 = 0xFF;   # w3-w101 99 bytes are all 1, so indicating free granules
my $grl = 0xF0;   # w102    last byte for the last 4 granules
#
my $format = 0;                           # format disk pack flag
my $dbuf="";                              # diskbuffer
my $yesno = "n";
my $header = "";                          # diskpack header info
my $pos;                                  # seek position
my $c;
my $i=0;

print "P860 X1210/X1215 disk pack premark, version $version\n";

my $file = shift (@ARGV);
if ($file eq "" || $file !~ /\.img/)
{
    print "No filename, or wrong .img extention\n";
    exit;
}

if( -e $file)
{
	open (IN,$file) || die "cannot open disk image file $file: $!";
	print "Verify X1210/X1215 moving head disk pack: $file\n";
}
else
{
	open (FDISK, ">:raw", $file) || die "cannot open image file $file: $!";
	print "Format X1210/X1215 moving head disk pack: $file\n";
	$format = 1;
}

if($format == 1)
{
	$cyl    = 0;
	$head   = 0;
	$sector = 0;
    $n      = 0;

	print "Writing the sector identifiers and initiate sectors\n";
	while($cyl < $cyls)
	{
		while($head < $heads)
		{
			print "$cyl\r";
			while($sector < $spt)
			{
				# format a logical sector in cylinder
				# print "Format sector $n s:$sector h:$head c:$cyl\n";
				formatlsector($sector,$head,$cyl);
				$n = $n + 1;
				$sector = $sector + 1;
			}
			$sector = 0;
			$head = $head + 1;
		}
		$head = 0;
		$cyl = $cyl + 1;
	}
	print "\n";
	print "Disk sector identifiers written to $n sectors.\n";

	while($yesno ne "y")
	{
		print "Provide the diskpack volume data\n"; 
		print "Disk pack label(16 chars, left adjusted)\n";
		$label = in(16);
		$label = UP($label);
		print "Date (dd/mm/yy)\n";
		$date = in(8);
		$date = UP($date);
		($day,$month,$year) = split(/\//,$date);
		print "Pack number (4 digits)\n";
		$packnbr = in(4);
		print "System UserId (8 chars, left adjusted)\n";
		$sysid = in(8);
		$sysid = UP($sysid);
		print "Label:        .$label.\n";
		print "Date:         .$day.$month.$year.\n";
		print "Pack Number:  .$packnbr.\n";
		print "System UserId:.$sysid.\n";
		print "Confirm with y if OK, n for retry\n";
		$yesno = in(1);
	}
	
	print "Write Volume Label\n";
	# write volume label
    $pos = VOLLAB;         # offset disk pack info in sector 0 (the volume label)
    $header=$vls1.$label.$vls2.$day.$vls3.$month.$vls4.$year.$vls5.$packnbr;
	#         12 +  16  + 10  + 2  +  2  +  2   +  2  +  2  +  18 +   4   = 70 bytes
	$i = 0;
	if(sysseek(FDISK, $pos, 0) ne undef)
	{	
	   while($i < 70)
	   {
			$c=unpack("C",substr($header,$i,1));
			wb($c);
			$i=$i+1;
	   }
	}
	else
	{
		print "Seek error while writing volume label\n";
	}
	# write the address of the physical sector with the badtrack list (2)
	ww(0x0002);
	# number of sectors per track
	ww(0x0010);
	# interlace factor
	ww(0x0003);
	# sector size in chars (incl heaeder trailer words: 410=/19A bytes
	ww(0x019A);
	# physical sector number of VTOC (5)
	ww(0x0005);
	# length of BITTAB in chars (102 = /66)
	ww(0x0066);
	print "Write Global Granule Allocation Table (BITTAB)\n";
    # write the initial granule table
    $pos = BITTAB;	# offset granule table in sector 0
	if(sysseek(FDISK, $pos, 0) ne undef)
	{	
		wb($gr0);
		wb($gr1);
	}
	else
	{
		print "Seek error while writing granule table\n";
	}
	$i=0;
	while($i<99)
	{
		wb($gr2);
		$i=$i+1;
	}
	wb($grl);     # last byte for the last 4 granules

	print "Write System Catalogue entry for the System UserId\n";
	# write the system catalogue entry
	# (1st entry of 8 words in logical sector 2 = physical sector 6)
	$pos = ($lsec * 2) * 6 + REC;      # seek to 1st catalogue entry
	my $cat = "CATALOG ";
	if(sysseek(FDISK, $pos, 0) ne undef)
	{	
		# write the catalogue identifier (entry 0)
		$i=0;
		while($i < length($cat))
		{
			$c=unpack("C",substr($cat,$i,1));
			wb($c);
			$i=$i+1;
		}		
		ww(0x0000);
		ww(0x0000);		
		ww(0x0000);
		ww(0x0000);		
		# write SystemId (4 words)
		$i=0;
		while($i < length($sysid))
		{
			$c=unpack("C",substr($sysid,$i,1));
			wb($c);
			$i=$i+1;
		}
		# write 2 * 0 (2 reserved words)
		ww(0x0000);
		ww(0x0000);
		# write a the sector pointer to the System Directory (sector 8)
		ww(0x0008);
		# write 0 ( 1reserved word)
		ww(0x0000);		
		#########
		# write End Of Catalogue Marker
		ww(0xFFFF);
	}
	else
	{
		print "Seek error while writing System catalogue entry\n";
	}	

	print "Write Monitor entry in System Directory\n";
	# write monitor directory entry in the Sytem Directory
	# the system Directory is in granule 1 and starts with logical sector 8
	# logical sector 8 = physical sector 8
	$pos = ($lsec * 2) * 8 + REC;      # seek to 1st directory entry
	my $dir = $sysid;
	if(sysseek(FDISK, $pos, 0) ne undef)
	{
		# write the directory identifier (entry 0)
		$i=0;
		while($i < length($dir))
		{
			$c=unpack("C",substr($dir,$i,1));
			wb($c);
			$i=$i+1;
		}		
		ww(0x0000);
		ww(0x0000);		
		ww(0x0000);
		ww(0x0000);
		# write the monitor dir entry in the sysdir
		# load module name (6 chars)
		ww(0x4D4F);  # MO
		ww(0x4E20);  # N<space>
		ww(0x2020);  # <space><space>
		# type: LM
		ww(0x4C4D);
		# 2 reserved words 2 * 0
		ww(0x0000);
		ww(0x0000);
		# pointer to sector where load module is stored (sector 16)
		ww(0x0010);
		# 1 reserved word
		ww(0x0000);
		#########
		# write End Of Directory Marker
		ww(0xFFFF);	
	}
	else
	{
		print "Seek error while writing Monitor directory entry\n";
	}	

    print "Write the granule allocation table for the monitor file area\n";
	# The monitor load module is stored in consecutive granules, starting
	# with granule 2 (logical sector /10 = physical sector /10)
	# This premark reserved 8 granules for the monitor (2..9)
	# Monitor code starts at logical sector 2 of the file and logical
	# sector 1 must contain the granule allocation table of the load module file.
	# This table is written here
	# The following entries are written for the 6 granules in the table:
	# granule 2 sector address /10 (cyl 0) == logical=physical sector 0 the the file
	# granule 3 sector address /18 (cyl 0)
	# granule 4 sector address /20 (cyl 1)
	# granule 5 sector address /28 (cyl 1)
	# granule 6 sector address /30 (cyl 1)
	# granule 7 sector address /38 (cyl 1)
	# granule 8 sector address /40 (cyl 2)
	# granule 9 sector address /48 (cyl 2)
	# the allocation table is in logical sector /11, physical sector /13
	$pos = ($lsec * 2) * 0x13 + REC;  # seek 1st entry in file allocation table	
	if(sysseek(FDISK, $pos, 0) ne undef)
	{	
		# write 8 entries in GRANTB of the lm file for the 8 granules
		ww(0x10);
		ww(0x18);
		ww(0x20);
		ww(0x28);
		ww(0x30);
		ww(0x38);
		ww(0x40);
		ww(0x48);
	}
	else
	{
		print "Seek error while writing Monitor file allocation table\n";
	}	
	
	close FDISK || die "can't close $file: $!";	
    ################################################################	
	# write IPL to physical sector 3
	my @cmd = (".\/ipl2disk.pl", "$file", "$iplfile");
	if(system (@cmd) != 0)
	{
		print "IPL load on disk failed\n";
	}
	else
	{
		print "IPL load on disk OK\n";
	}
		
	################################################################
	print "Verify the sector identifiers.\n";
	open (FDISK,$file) || die "cannot open disk image file $file: $!";
	$i=0;
	while($i < $n)
	{
		readsector($i); # $i is a logical disk sector number (0..$n)
		$i=$i+1;
	}
	print "$i sector identifiers verified.\n";
	close FDISK || die "can't close $file: $!";	
}
else
{
	print "Verify the sector identifiers.\n";
	open (FDISK,$file) || die "cannot open disk image file $file: $!";
	$i=0;
	while($i < $n)
	{
		readsector($i); # $i is a logical disk sector number (0..$n)
		$i=$i+1;
	}
	print "$i sector identifiers verified.\n";
	close (FDISK) || die "can't close $file: $!";
}

print "Disk Pack initialization ready\n";

####################################################################

# read character string of requested length from console
sub in {
	my $l = shift;
	my $i = 0;
	my $s = "";
	while($i != $l)
	{
		print "? ";
		$s = <STDIN>;
		chomp($s);
		$i = length($s);
	}
	return $s;
}

# read a logical disk sector into buffer dbuf (0..6495)
sub readsector {
	my $s = shift;                 # logical disk sector number
    my $c = int($s/($spt*$heads)); # cylinder
	my $r = $s%($spt*$heads);
	my $h = int($r/$spt);          # head
	my $l = $r%$spt;               # logical sector
	my $p = secphys($l);           # physical sector
	my $w1;
	# print "S:$s C:$c H:$h L:$l PS:$p\n";
	$pos = $c*$spt*$heads*$lsec*2 + $h*$spt*$lsec*2 + $p*$lsec*2;
	if(sysseek(FDISK, $pos, 0) ne undef)
	{	
        if(sysread(FDISK, $dbuf, $lsec*2) != $lsec*2)
		{
			print "Disk read error while reading sector $s\n";
		}
		# check the sector identifier
		$w1 = unpack("C", substr($dbuf,0,1)) << 8;
		$w1 = $w1 | unpack("C", substr($dbuf,1,1));
		if($SEC_IDENT == TEST)    # phys sec number + cyl number
		{
			my $dc = $w1 & 0x1FF;
			my $dh = $w1 >> 15;
			my $dp = ($w1 >> 11) & 0xF;
			if($c == $dc && $h == $dh && $p == $dp)
			{
				print "$s\r";
			}
			else
			{
				print "Disk format error: requested S:$s C:$c H:$h L:$l P:$p\n";
				print "from disk provided identifier :   C:$dc H:$dh P:$dp\n";
			}
		}
		elsif($SEC_IDENT == P800) # only cylinder number
		{
			if($c == $w1)
			{
				print "$s\r";
			}
			else
			{
				print "Disk format error: requested S:$s C:$c\n";
				print "from disk provided identifier :   C:$w1\n";
			}
		}
		else
		{
			print "ERROR: Sector identifier format selection\n";
			exit;
		}			
	}
	else
	{
		print "Seek error while reading sector $s\n";
	}	
}

# format a logical sector in a track (0..15)
sub formatlsector {
	my $sl = shift;
	my $sf = secphys($sl); # physical sector number
	my $h  = shift;        # head number
	my $c  = shift;        # cylinder number
	my $w1;
	my $w;
	my $i;
	$pos = $c*$spt*$heads*$lsec*2 + $h*$spt*$lsec*2 + $sf*$lsec*2;
	# print "Physical sector $sf position: $pos\n";
	if(sysseek(FDISK, $pos, 0) ne undef)
	{
		if($SEC_IDENT == TEST)
		{
			# create sector identifier (head..physical sector number..cylinder number)
			$w1 = $h << 15;
			$w1 = $w1 | (($sf & 0xF) << 11);
			$w1 = $w1 | ($c & 0x1FF);
		}
		elsif($SEC_IDENT == P800)
		{
			$w1 = $c & 0x1FF;
		}
		else
		{
			print "ERROR: Sector identifier format selection\n";
			exit;
		}
		ww($w1); # write word 1 (sector identifier
		$w = 0;
		$i = 1;
		# write the 204 other words in the sector
		while($i < $lsec)
		{
			ww($w);
			$i=$i+1;
		}
	}
	else
	{
		print "Seek error, the formatted disk might be wrong\n";
	}
}

# write next word to disk
sub ww {
	my $w = shift;
	wb($w >> 8);
	wb($w & 0xFF);
}

# write next byte to disk
sub wb {
	my $b =shift;
	$b = pack("C", $b);
	if(syswrite(FDISK, $b, 1) != 1)
	{
		print "Disk write error\n";
	}
}

# convert a logical sector number in track to a physical sector number in track
sub secphys {
	my $sl = shift;
	$sl=$sl*3;
	if($sl > 16)
	{
		$sl=$sl-16;
		if($sl > 16)
		{
			$sl=$sl-16;
		}
	}
	return $sl;
}

# translate string to upper case
sub UP {
	my $s = shift;
    $_ = $s;
	tr/a-z/A-Z/;
	return $s = $_;
}

